//---------------------------------------------------------------------------

//#pragma hdrstop

//---------------------------------------------------------------------------

//#pragma argsused

#include <stdio.h>
#include <stdint.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>

#include "defs.h"
#include "print.h"
#include "serial.h"
#include "avroo.h"

const char* portname 		= NULL; //COM_DEFAULT;
const char* portconfig 		= NULL; //CFG_DEFAULT;
const char* trigger 		= "<rest>";
const char* flashfile 		= NULL;
const char* eepfile 		= NULL;
bool keepalive				= false;
unsigned long timeout		= BOOTLOADER_CATCH_TIMEOUT_MS;

void usage(const char* progname)
{
	printf_yellow("usage: \r\n");
	printf("%s [Options]\r\n", progname);
	printf("  -p COMPORT     e.g. " COM_DEFAULT " ... \r\n");
	printf("  -c PORTCONFIG  e.g. \"" CFG_DEFAULT "\" \r\n");
	printf("  -f FLASHFILE   Intel Hex (.hex) or Binary Image File (.bin) \r\n");
	printf("  -e EEPROMFILE  Intel Hex (.hex or .eep) or Binary Image File (.bin) \r\n");
	printf("  -t TRIGGER     Triggerstring for the Bootloader to launch. \r\n");
	printf("  -k             Keep Bootloader alive until keypress after successful upload. \r\n");
	printf("  -o TIMEOUT     Timeout in seconds to catch Bootloader. Use 0 for infinite. \r\n");
}

bool scanargs(int argc, char** argv)
{
	int argp = 1;

	for (int i = 1; i < argc; i++)
	{
		char* arg = argv[i];

		if (STREQU(arg, "-p") || STREQU(arg, "--portname"))
		{
			portname = argv[i+1];
			i++;
			argp += 2;
		}
		else if (STREQU(arg, "-c") || STREQU(arg, "--portconfig"))
		{
			portconfig = argv[i+1];
			i++;
			argp += 2;
		}
		else if (STREQU(arg, "-t") || STREQU(arg, "--trigger"))
		{
			trigger = argv[i+1];
			i++;
			argp += 2;
		}
		else if (STREQU(arg, "-f") || STREQU(arg, "--flashfile"))
		{
			flashfile = argv[i+1];
			i++;
			argp += 2;
		}
		else if (STREQU(arg, "-e") || STREQU(arg, "--eepfile"))
		{
			eepfile = argv[i+1];
			i++;
			argp += 2;
		}
		else if (STREQU(arg, "-o") || STREQU(arg, "--timeout"))
		{
			timeout = atol(argv[i+1]) * 1000;
			i++;
			argp += 2;
		}
		else if (STREQU(arg, "-k") || STREQU(arg, "--keepalive"))
		{
			keepalive = true;
			argp += 1;
		}
		else if (STREQU(arg, "-xbf") || STREQU(arg, "--transferbufsize-flash"))
		{
			g_transferbufsize_flash = atoi(argv[i+1]);
			i++;
			argp += 2;
		}
		else if (STREQU(arg, "-xbe") || STREQU(arg, "--transferbufsize-eeprom"))
		{
			g_transferbufsize_eeprom = atoi(argv[i+1]);
			i++;
			argp += 2;
		}
		else if (STREQU(arg, "-h") || STREQU(arg, "--help"))
		{
			usage(argv[0]);
			exit(0);
			return false;
		}
		else if (STREQU(arg, "-v") || STREQU(arg, "--version"))
		{
			printf("Version: " COMPILE_TIMESTAMP "\r\n");
			exit(0);
			return false;
		}		
		else
		{
			eprintf("Error: Unrecognized command line option \"%s\".\r\n", arg);
			usage(argv[0]);
			return false;
		}
	}

	return (argc == argp);
}

bool query_comportname(void)
{
	static char cpn_buf[FILENAME_MAX];

	printf("Serial Ports in System:\r\n");
	serial_enumports();
	printf("Select Serial Port: ");
	fgets(cpn_buf, sizeof(cpn_buf), stdin);



	#ifdef WIN32
	if (isdigit(*cpn_buf))
	{
		int num = atoi(cpn_buf);
		snprintf(cpn_buf, sizeof(cpn_buf), "COM%u", num);
	}
	#endif

	portname = strtrim(cpn_buf);
}

int main(int argc, char* argv[])
{
	printf_init();

#if (defined _WIN32)
	printf("avrootloader (_WIN32): start...\r\n");
#elif (defined LINUX)
	printf("avrootloader (LINUX): start...\r\n");
#endif

	if (!scanargs(argc, argv))
	{
		return 1;
	}

	if (g_configpath == NULL)
	{
		g_configpath = argv[0];
		char* last_delim = strrchr(g_configpath, PATH_DELIMITER);
		if (last_delim)
			*last_delim = '\0';
	}

	if (portname == NULL)
	{
		query_comportname();
	}

	if (portconfig == NULL)
	{
		static char cpc_buf[64];
		printf("Enter serial port config: ");
		fgets(cpc_buf, sizeof(cpc_buf), stdin);
		portconfig = cpc_buf;
	}

	if (flashfile == NULL)
	{
		static char ffn_buf[FILENAME_MAX];
		flashfile = openfiledialog("Select Flash File", ffn_buf, sizeof(ffn_buf));
	}

	if (eepfile == NULL)
	{
		static char efn_buf[FILENAME_MAX];
		eepfile = openfiledialog("Select EEPROM File", efn_buf, sizeof(efn_buf));
	}

	printf("avrootloader - upload settings: \r\n");	
	printf("- Portname  : %s \r\n", portname);
	printf("- Portconfig: %s \r\n", portconfig);
	printf("- Flashfile : %s \r\n", flashfile);	
	printf("- Eepromfile: %s \r\n", eepfile);		
	printf("- Trigger   : %s \r\n", trigger);

	bool res = avroo_upload(
		portname, portconfig,
		trigger, strlen(trigger),
		flashfile,
		eepfile,
		keepalive,
		timeout
	);

	//waitkeypress("Hit any Key to exit...\r\n");

	printf_deinit();

	return res ? 0 : 1;
}
