//---------------------------------------------------------------------------

#ifndef printH
#define printH
//---------------------------------------------------------------------------

#include <stdio.h>
#include <time.h>
#include <stdarg.h>




#if (defined _WIN32)
	#include <conio.h> // for kbhit
	#include <windows.h>
	#define CONSOLE_COLOR_DARKGREEN			( FOREGROUND_GREEN )
	#define CONSOLE_COLOR_BRIGHTGREEN		( FOREGROUND_GREEN | FOREGROUND_INTENSITY )
	#define CONSOLE_COLOR_DARKRED			( FOREGROUND_RED )
	#define CONSOLE_COLOR_BRIGHTRED			( FOREGROUND_RED | FOREGROUND_INTENSITY )
	#define CONSOLE_COLOR_DARKBLUE			( FOREGROUND_BLUE )
	#define CONSOLE_COLOR_BRIGHTBLUE		( FOREGROUND_BLUE | FOREGROUND_INTENSITY )
	#define CONSOLE_COLOR_YELLOW			( FOREGROUND_GREEN | FOREGROUND_RED | FOREGROUND_INTENSITY )	
	#define CONSOLE_COLOR_NORMAL			( FOREGROUND_BLUE | FOREGROUND_GREEN | FOREGROUND_RED )
	
	#define printf_red(...)					printf_color(CONSOLE_COLOR_BRIGHTRED, 	__VA_ARGS__)
	#define printf_grn(...)					printf_color(CONSOLE_COLOR_BRIGHTGREEN, __VA_ARGS__)
	#define printf_green(...)				printf_color(CONSOLE_COLOR_BRIGHTGREEN, __VA_ARGS__)
	#define printf_blu(...)					printf_color(CONSOLE_COLOR_BRIGHTBLUE, 	__VA_ARGS__)
	#define printf_blue(...)				printf_color(CONSOLE_COLOR_BRIGHTBLUE, 	__VA_ARGS__)
	#define printf_yel(...)					printf_color(CONSOLE_COLOR_YELLOW, 		__VA_ARGS__)
	#define printf_yellow(...)				printf_color(CONSOLE_COLOR_YELLOW, 		__VA_ARGS__)
	
	void printf_init(void);
	void printf_color(int color, const char *fmt, ...);
	void printf_deinit(void);
	
	#define sleep_ms(MS)					Sleep(MS)
	#define GetTickCount_ms()				GetTickCount()
	#define GetLastErrorText()				strerror(GetLastError())
#endif

#if (defined LINUX)
	#include <errno.h>
	#include <string.h>

	#define CONSOLE_COLOR_BRIGHTGREEN		"\x1b[32m"
	#define CONSOLE_COLOR_BRIGHTRED			"\x1b[31m"
	#define CONSOLE_COLOR_BRIGHTBLUE		"\x1b[34m"
	#define CONSOLE_COLOR_YELLOW			"\x1b[33m"
	#define CONSOLE_COLOR_NORMAL			"\x1b[0m"

	#define printf_red(format, ...)			printf(CONSOLE_COLOR_BRIGHTRED 		format CONSOLE_COLOR_NORMAL, ##__VA_ARGS__)
	#define printf_grn(format, ...)			printf(CONSOLE_COLOR_BRIGHTGREEN 	format CONSOLE_COLOR_NORMAL, ##__VA_ARGS__)
	#define printf_green(format, ...)		printf(CONSOLE_COLOR_BRIGHTGREEN 	format CONSOLE_COLOR_NORMAL, ##__VA_ARGS__)
	#define printf_blu(format, ...)			printf(CONSOLE_COLOR_BRIGHTBLUE 	format CONSOLE_COLOR_NORMAL, ##__VA_ARGS__)
	#define printf_blue(format, ...)		printf(CONSOLE_COLOR_BRIGHTBLUE 	format CONSOLE_COLOR_NORMAL, ##__VA_ARGS__)
	#define printf_yel(format, ...)			printf(CONSOLE_COLOR_YELLOW 		format CONSOLE_COLOR_NORMAL, ##__VA_ARGS__)
	#define printf_yellow(format, ...)		printf(CONSOLE_COLOR_YELLOW 		format CONSOLE_COLOR_NORMAL, ##__VA_ARGS__)
	
	void printf_init(void);
	#define printf_deinit()
	
	void sleep_ms(unsigned int ms);
	unsigned long GetTickCount_ms(void);
	int kbhit(void);
	#define GetLastError()				errno
	#define getch()						getchar()
	#define GetLastErrorText()			strerror(errno)
#endif

// http://gcc.gnu.org/onlinedocs/cpp/Variadic-Macros.html
//#define eprintf(...) 					fprintf(stderr, __VA_ARGS__)
//#define eprintf(format, ...)			fprintf (stderr, format, __VA_ARGS__)
//#define eprintf(format, ...)			fprintf (stderr, format, ##__VA_ARGS__) 
#define eprintf(...) 					printf_red(__VA_ARGS__)



#define HEXDUMP_LINELEN	16
void hexdump(unsigned int addr, const char* buf, int len, FILE* file=stdout);
// void perror(const char* msg);
void waitkeypress(const char* prompt);
char* timestr(time_t timestamp);
#define nowstr()	timestr(time(NULL))

char* strtrim(char* str);

//void set_title(const char *fmt, ...);

void printhex(const char* buf, int len);
void printstr(const char* buf, int len);
char* openfiledialog(const char* prompt, char* buf, int len);
char wheel(void);

#endif
