//---------------------------------------------------------------------------

// #pragma hdrstop

#include "serial.h"
#include "print.h"

//---------------------------------------------------------------------------

#include <stdio.h>

#if (defined _WIN32)

// based on
// http://msdn.microsoft.com/en-us/library/ff802693.aspx

#include <windows.h>

HANDLE hFile = INVALID_HANDLE_VALUE;

#define rx_dump(BUF, LEN) //xferdump(XFERDIR_RX, BUF, LEN)
#define tx_dump(BUF, LEN) //xferdump(XFERDIR_TX, BUF, LEN)

void serial_enumports(void)
{
	#define REGPATH_SERIALPORTS	"HARDWARE\\DEVICEMAP\\SERIALCOMM" // in HKEY_LOCAL_MACHINE

	//vector<string> vs;

	HKEY hKey;

	if (RegOpenKey(HKEY_LOCAL_MACHINE, REGPATH_SERIALPORTS, &hKey) == ERROR_SUCCESS)
	{
		DWORD cValues;
		DWORD retCode;

		retCode = RegQueryInfoKey(hKey, NULL, NULL, NULL, NULL, NULL, NULL, &cValues, NULL, NULL, NULL, NULL);

		if ((retCode == ERROR_SUCCESS) && cValues)
		{
			for (unsigned int i=0, retCode=ERROR_SUCCESS; i<cValues; i++)
			{
				char ValName[96];
				ValName[0] = '\0';
				DWORD ValNameLen = sizeof(ValName) - 1;
				DWORD dwType = REG_NONE;

				char ValData[32];
				ValData[0] = '\0';
				DWORD ValDataLen = sizeof(ValData) - 1;

				retCode = RegEnumValue(hKey, i, ValName, &ValNameLen, NULL, &dwType, ValData, &ValDataLen);

				if ((retCode == ERROR_SUCCESS) && (dwType == REG_SZ))
				{
					//printf("  %s\t(%s)\n", ValData, ValName);
					printf("%s\n", ValData);
					//vs.push_back(string(ValData));
				}
			}
		}

		RegCloseKey(hKey);
	}

	//return vs;
}

bool serial_config(const char* cfgstr)
{
	DCB dcb;

	if (cfgstr == NULL) return true;

	if (hFile == INVALID_HANDLE_VALUE)
	{
		return false;
    }

	if (!GetCommState(hFile, &dcb))
	{
		perror("SetCommState() failed.");
		return false;
	}

	if (!BuildCommDCB(cfgstr, &dcb))
	{
		perror("BuildCommDCB() failed.");
		return false;
	}

	if (!SetCommState(hFile, &dcb))
	{
		perror("SetCommState() failed.");
        return false;
	}

	return true;
}

bool serial_settimeout(int timeout_ms)
{
	// timeouts:
	COMMTIMEOUTS timeouts;

	if (!GetCommTimeouts(hFile, &timeouts))
	{
		perror("GetCommTimeouts() failed.");
		CloseHandle(hFile);
		return false;
	}

	timeouts.ReadIntervalTimeout 			= MAXDWORD;
	timeouts.ReadTotalTimeoutMultiplier 	= MAXDWORD;
	timeouts.ReadTotalTimeoutConstant 		= timeout_ms; // MAXDWORD - 1;
	timeouts.ReadIntervalTimeout 			= MAXDWORD;
	timeouts.WriteTotalTimeoutMultiplier 	= 0;
	timeouts.WriteTotalTimeoutConstant 		= 0;

	if (!SetCommTimeouts(hFile, &timeouts))
	{
		perror("SetCommTimeouts() failed.");
		CloseHandle(hFile);
		return false;
	}

	return true;
}

void serial_flush(void)
{
	PurgeComm(
		hFile,
		PURGE_RXABORT | PURGE_RXCLEAR | PURGE_TXABORT | PURGE_TXCLEAR
	);
}

bool serial_open(const char* portname, const char* cfgstr)
{
	if (hFile != INVALID_HANDLE_VALUE)
	{
		printf("Error: serial port already open.\r\n");
		return false;
	}

	char _portname[32];
	snprintf(_portname, sizeof(_portname), "\\\\.\\%s", portname);

	hFile = CreateFile(
		_portname,
		GENERIC_READ | GENERIC_WRITE,
		0,
		NULL,
		OPEN_EXISTING,
		0, //FILE_FLAG_OVERLAPPED,
		NULL
	);

	if (INVALID_HANDLE_VALUE == hFile)
	{
		eprintf("Error: opening serial port %s failed: %s \r\n", portname, GetLastErrorText());
		return false;
	}

	serial_config(cfgstr);

	//serial_settimeout(500);

	if (!SetCommMask(hFile, EV_RXCHAR))
	{
		perror("SetCommMask(EV_RXCHAR) failed.");
		return false;
	}

	serial_flush();

	printf("Opened serial port %s (%s).\r\n", portname, cfgstr);
	return true;

}

bool serial_close(void)
{
	if (hFile == INVALID_HANDLE_VALUE)
	{
		eprintf("Error: serial port already closed.\r\n");
		return false;
	}

	CloseHandle(hFile);
	hFile = INVALID_HANDLE_VALUE;

	printf("Closed serial port.\r\n");
	return true;
}

int serial_puts(const char* buf, int len)
{
	len = len ? len : strlen(buf);

	DWORD txcnt = 0;
	bool res = WriteFile(hFile, buf, len, &txcnt, NULL); // &ovlWrite)

	if (!res)
	{
		perror("WriteFile() failed.");
	}
	else
	{
		tx_dump(buf, txcnt);
	}

	return res ? txcnt : -1;
}

int serial_putc(char c)
{
	return serial_puts(&(c), sizeof(char));
}

int serial_gets(char* buf, int len, int timeout)
{
	DWORD rxcnt = 0;

	serial_settimeout(timeout);

	bool res = ReadFile(hFile, buf, len, &rxcnt, NULL); // &ovlRead)

	if (!res)
	{
		perror("ReadFile() failed.");
	}
	else
	{
		rx_dump(buf, rxcnt);
	}

	return res ? rxcnt : -1;
}

#endif 

#if (defined LINUX)

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/select.h>
#include <termios.h>
#include <sys/ioctl.h>

#define INVALID_FIELNO			(-1)
#define INVALID_SOCKET			(-1)
#define SOCKET_ERROR			(-1)

#define VALIDSOCKET(S)			((S) != INVALID_SOCKET)
#define VALIDFILENO(F)			((F) != INVALID_FIELNO)
#define INVALIDSOCKET(S)		((S) == INVALID_SOCKET)
#define INVALIDFILENO(F)		((F) == INVALID_FIELNO)

int g_tty_fileno			= INVALID_FIELNO;

void serial_enumports(void)
{
	printf_red("TODO: list all /dev/tty/S* /dev/ttyUSB* /dev/ttyAMA* and similars ...\r\n");
}

// cfgstr example: "baud=115200 parity=N data=8 stop=1 dtr=off rts=off"
// ... equivalent to ...
// $ stty -F /dev/ttyUSB1 ispeed 115200 ospeed 115200 cs8 -crtscts -echo raw
bool serial_config(const char* cfgstr)
{
	if (cfgstr == NULL) return true;

	struct termios options;
	if (tcgetattr(g_tty_fileno, &options))
	{
		eprintf("Error: getting terminal settings (tcgetattr) failed: %s\r\n", GetLastErrorText());
		return false;
	}	

	cfmakeraw(&options);
    options.c_cflag |= (CLOCAL | CREAD);    		// Enable the receiver and set local mode
    options.c_cflag &= ~CRTSCTS;            		// Disable hardware flow control  
    options.c_lflag &= ~(ICANON | ECHO | ISIG);		// Enable data to be processed as raw input
	
	const char* delims = " \t\r\n";	
	char* _cfgstr = strdup(cfgstr);
	char* tok = strtok(_cfgstr, delims);
	while (tok != NULL) 
	{
		//printf("Assigning: %s ... \r\n", tok);
		
		if STRSTARTS(tok, "baud=")
		{
			unsigned long baud = atol(tok + LLEN("baud="));
			const unsigned long baud2cfspeed[] =
			{
				0		, B0     ,
				50		, B50    ,
				75		, B75    ,
				110		, B110   ,
				134		, B134   ,
				150		, B150   ,
				200		, B200   ,
				300		, B300   ,
				600		, B600   ,
				1200	, B1200  ,
				1800	, B1800  ,
				2400	, B2400  ,
				4800	, B4800  ,
				9600	, B9600  ,
				19200	, B19200 ,
				38400	, B38400 ,
				57600	, B57600 ,
				115200	, B115200,
				230400	, B230400,
			};
			for (int i=0; i<ELEMCNT(baud2cfspeed); i+=2)
			{
				if (baud == baud2cfspeed[i])
				{
					unsigned long cfspeed = baud2cfspeed[i+1];
					cfsetispeed(&options, cfspeed);
					cfsetospeed(&options, cfspeed);
					break;
				}
			}
		}
		else if STRSTARTS(tok, "parity=")
		{
			char parity = *(tok + LLEN("parity="));
			
			options.c_cflag &= ~(PARENB | PARODD); // parity disabled
			
			switch (parity)
			{
				case 'E':
				{
					options.c_cflag |= (PARENB);
					break;
				}
				case 'O':
				{
					options.c_cflag |= (PARENB | PARODD);
					break;
				}
				// todo: Mark, Space
			}
		}	
		else if STRSTARTS(tok, "data=")
		{
			int data = atoi(tok + LLEN("data="));
			
			const int data2cs[] = {
				5, CS5, 
				6, CS6, 
				7, CS7, 
				8, CS8,
			};
			
			for (int i=0; i<ELEMCNT(data2cs); i+=2)
			{
				if (data == data2cs[i])
				{
					int cs = data2cs[i+1];
					options.c_cflag = (options.c_cflag & ~CSIZE) | cs;
					break;
				}
			}
		}
		else if STRSTARTS(tok, "stop=")
		{
			int stop = atoi(tok + LLEN("stop="));
			options.c_cflag &= ~CSTOPB;
			
			if (stop > 1)
				options.c_cflag |= CSTOPB;
		}	
		else if STRSTARTS(tok, "dtr=")
		{
			bool val = STREQU(tok + LLEN("dtr="), "on");
			
			int status;

			if (ioctl(g_tty_fileno, TIOCMGET, &status))
				eprintf("Error: ioctl (getting rts) failed: %s\r\n", GetLastErrorText());			
			
			if (val)
				status |= TIOCM_DTR;
			else
				status &= ~TIOCM_DTR;

			if (ioctl(g_tty_fileno, TIOCMSET, &status))
				eprintf("Error: ioctl (setting rts) failed: %s\r\n", GetLastErrorText());
			
		}			
		else if STRSTARTS(tok, "rts=")
		{
			bool val = STREQU(tok + LLEN("rts="), "on");
			
			int status;

			if (ioctl(g_tty_fileno, TIOCMGET, &status))
				eprintf("Error: ioctl (getting rts) failed: %s\r\n", GetLastErrorText());			
			
			if (val)
				status |= TIOCM_RTS;
			else
				status &= ~TIOCM_RTS;

			if (ioctl(g_tty_fileno, TIOCMSET, &status))
				eprintf("Error: ioctl (setting rts) failed: %s\r\n", GetLastErrorText());
		}
		else
		{
			eprintf("Error: unrecognized option \"%s\" in port configuration.", tok);
		}
		
		tok = strtok(NULL, delims);
	}
	free(_cfgstr);	
	
	if (tcsetattr(g_tty_fileno, TCSANOW, &options))
	{
		eprintf("Error: setting terminal settings (tcsetattr) failed: %s\r\n", GetLastErrorText());
		return false;
	}

	return true;
}

void serial_flush(void)
{
	tcflush(g_tty_fileno, TCIOFLUSH);
}

bool serial_open(const char* portname, const char* cfgstr)
{
	g_tty_fileno = open(portname, O_RDWR | O_NOCTTY);
	
	if INVALIDFILENO(g_tty_fileno)
	{
		eprintf("Error: opening serial port %s failed: %s \r\n", portname, GetLastErrorText());
		return false;
	}
	
	serial_config(cfgstr);
	
	printf("Opened serial port %s (%s) handle=%d.\r\n", portname, cfgstr, g_tty_fileno);
	
	return true;
}

int  serial_puts(const char* buf, int len)
{
	len = len ? len : strlen(buf);
	int n = write(g_tty_fileno, buf, len);
	return n;
}

int  serial_putc(char c)
{
	return serial_puts(&(c), sizeof(char));
}

int  serial_gets(char* buf, int len, int timeout)
{
	struct timeval tv = {timeout/1000, (timeout%1000)*1000};

	fd_set fd_read;
	FD_ZERO(&fd_read);
	FD_SET(g_tty_fileno ,&fd_read);

	int n = select(g_tty_fileno + 1, &fd_read, NULL, NULL, &tv);

	if ((n > 0) && FD_ISSET(g_tty_fileno, &fd_read))
	{
		return read(g_tty_fileno, buf, len);	
	}
	else if (n == 0)
	{
		// printf("Timeout\n");
		return 0;
	}
	else
	{
		eprintf("Error: reading serial port (select) failed: %s\r\n", GetLastErrorText());
		return n;
	}

	return n;
}

bool serial_close(void)
{
	if INVALIDFILENO(g_tty_fileno)
	{
		eprintf("Error: serial port already closed.\r\n");
		return false;
	}

	close(g_tty_fileno);
	g_tty_fileno = INVALID_FIELNO;

	printf("Closed serial port.\r\n");
	
	return true;
}

#endif

/*
enum {XFERDIR_RX, XFERDIR_TX, XFERDIR_END};

void xferdump(int xferdir, const char* buf, int len)
{
	static int last_xferdir = XFERDIR_END;
	static int count[XFERDIR_END] = {0, 0};

	if (last_xferdir != xferdir)
	{
		count[xferdir] = 0;
		printf((xferdir == XFERDIR_RX) ? "Rx:\r\n" : "Tx:\r\n");
	}

	hexdump(count[xferdir], buf, len);
	count[xferdir] += len;

	last_xferdir = xferdir;
}
*/

#include <string.h>

int serial_readline(char* buf, int len, int timeout, char endl)
{
	memset(buf, '\0', len);

	int te = GetTickCount_ms() + timeout;

	for (int i=0; i<len; i++)
	{
		if (GetTickCount_ms() > te)
		{
			return 0;
		}

		char c;
		int r = serial_gets(&c, sizeof(c), timeout);

		if (r > 0)
		{
			buf[i] = c;
		}

		if (c == endl)
		{
			return i + 1;
		}
	}
}